import pandas as pd
import numpy as np
import datetime as datetime
from ecopipeline import ConfigManager
import re
import mysql.connector.errors as mysqlerrors
from datetime import timedelta
from ecopipeline.event_tracking.Alarm import Alarm

class HPWHOutage(Alarm):
    """
    Detects possible heat pump failures or outages by checking if heat pump power consumption falls below
    an expected ratio of total system power over a rolling period, or by checking for non-zero values in
    a direct alarm variable from the heat pump controller.

    Variable_Names.csv configuration:
      alarm_codes column: HPOUTGE or HPOUTGE:### where ### provides the bound for the variable (see types below).
      variable_name column: determines the role of the variable by its first underscore-separated part:
        PowerIn_[name] - HP power variable. Bound (###) from alarm_codes is the minimum expected ratio of HP
            power to total system power over the rolling period (default 0.3 for 30%). Alarm triggers when
            HP power falls below this ratio. Must be in the same units as the PowerIn_Total variable.
        PowerIn_Total - Total system power variable. No bound needed in alarm_codes (use just HPOUTGE).
            Required when using PowerIn variables.
        Alarm_[name] - Direct alarm variable from HP controller. No bound needed in alarm_codes (use just HPOUTGE).
            Alarm triggers on any non-zero value. If triggered, the ratio check is skipped.

    Parameters
    ----------
    day_table_name : str
        Name of the site's daily aggregate value table in the database for fetching historical data.
    default_power_ratio : float
        Default minimum ratio threshold for PowerIn variables when no bound is specified (default 0.3).
        Alarm triggers when HP power falls below this ratio of total power over the rolling period.
    ratio_period_days : int
        Number of days to use for the rolling power ratio calculation (default 7). Must be greater than 1.
    """
    def __init__(self, bounds_df : pd.DataFrame, day_table_name : str, default_power_ratio : float = 0.3,
                   ratio_period_days : int = 7):
        alarm_tag = 'HPOUTGE'
        type_default_dict = {
                'PowerIn': default_power_ratio,
                'PowerIn_Total': None,
                'Alarm': None
            }
        self.day_table_name = day_table_name # TODO this could be a security issue. Swap it for config manager
        self.default_power_ratio = default_power_ratio
        self.ratio_period_days = ratio_period_days
        super().__init__(bounds_df, alarm_tag,type_default_dict, two_part_tag = True)

    def specific_alarm_function(self, df: pd.DataFrame, daily_df : pd.DataFrame, config : ConfigManager):
        for alarm_id in self.bounds_df['alarm_code_id'].unique():
            alarm_triggered = False
            id_group = self.bounds_df[(self.bounds_df['alarm_code_id'] == alarm_id) | (self.bounds_df['alarm_code_type'] == 'PowerIn_Total')]
            # Get T and SP alarm codes for this ID
            pow_codes = id_group[id_group['alarm_code_type'] == 'PowerIn']
            tp_codes = id_group[id_group['alarm_code_type'] == 'PowerIn_Total']
            alrm_codes = id_group[id_group['alarm_code_type'] == 'Alarm']
            if len(alrm_codes) > 0:
                for i in range(len(alrm_codes)):
                    alrm_var_name = alrm_codes.iloc[i]['variable_name']
                    alrm_pretty_name = alrm_codes.iloc[i]['pretty_name']
                    if alrm_var_name in df.columns:
                        for day in daily_df.index:
                            next_day = day + pd.Timedelta(days=1)
                            filtered_df = df.loc[(df.index >= day) & (df.index < next_day)]
                            if not filtered_df.empty:
                                # Find all consecutive blocks where alarm variable is non-zero
                                alarm_mask = filtered_df[alrm_var_name] != 0
                                if alarm_mask.any():
                                    # Find consecutive groups
                                    group = (alarm_mask != alarm_mask.shift()).cumsum()

                                    # Iterate through each consecutive block of non-zero values
                                    for group_id in alarm_mask.groupby(group).first()[lambda x: x].index:
                                        streak_indices = alarm_mask[group == group_id].index
                                        start_time = streak_indices[0]
                                        end_time = streak_indices[-1]
                                        streak_length = len(streak_indices)
                                        alarm_value = filtered_df.loc[start_time, alrm_var_name]

                                        self._add_an_alarm(start_time, end_time, alrm_var_name,
                                            f"Heat pump alarm triggered: {alrm_pretty_name} was {alarm_value} for {streak_length} minutes starting at {start_time}.")
                                        alarm_triggered = True
            elif len(pow_codes) > 0 and len(tp_codes) != 1:
                raise Exception(f"Improper alarm codes for heat pump outage with id {alarm_id}. Requires 1 total power (TP) variable.")
            if len(pow_codes) > 0 and len(tp_codes) == 1 and not alarm_triggered:
                if self.ratio_period_days <= 1:
                    print("HP Outage alarm period, ratio_period_days, must be more than 1")
                else: 
                    tp_var_name = tp_codes.iloc[0]['variable_name'] 
                    daily_df_copy = daily_df.copy()
                    daily_df_copy = self._append_previous_days_to_df(daily_df_copy, config, self.ratio_period_days, self.day_table_name)
                    for i in range(self.ratio_period_days - 1, len(daily_df_copy)):
                        start_idx = i - self.ratio_period_days + 1
                        end_idx = i + 1
                        day = daily_df_copy.index[i]
                        block_data = daily_df_copy.iloc[start_idx:end_idx].sum()
                        for j in range(len(pow_codes)):
                            pow_var_name = pow_codes.iloc[j]['variable_name']
                            pow_var_bound = pow_codes.iloc[j]['bound']
                            if block_data[pow_var_name] < block_data[tp_var_name] * pow_var_bound:
                                self._add_an_alarm(day, day + timedelta(1), pow_var_name, f"Possible Heat Pump failure or outage.", False,
                                                   certainty='med')


import pandas as pd
import numpy as np
import datetime as datetime
from ecopipeline import ConfigManager
import re
import mysql.connector.errors as mysqlerrors
from datetime import timedelta
from ecopipeline.event_tracking.Alarm import Alarm

class BlownFuse(Alarm):
    """
    Detects blown fuse conditions for heating elements by identifying when an element is drawing power
    but significantly less than its expected draw, suggesting a blown fuse.

    Variable_Names.csv configuration:
      alarm_codes column: BLWNFSE:### where ### is the expected kW draw when the element is fully on.
      variable_name column: Must start with PowerIn_ (e.g., PowerIn_ERElement1).
        PowerIn_[name] - Element power variable. Bound (###) from alarm_codes is the expected kW draw (default 30 kW).
            Alarm triggers when element is on (power > default_power_threshold) but drawing less than
            (expected_draw - default_power_range) for fault_time consecutive minutes.

    Parameters
    ----------
    default_power_threshold : float
        Minimum power level (kW) to consider the element 'on' (default 1.0).
    default_power_range : float
        Allowable variance below expected power draw (default 2.0). Alarm triggers when actual draw < (expected - range).
    default_power_draw : float
        Default expected power draw in kW when no bound is specified in the alarm code (default 30).
    fault_time : int
        Number of consecutive minutes the fault condition must persist before triggering an alarm (default 3).
    """
    def __init__(self, bounds_df : pd.DataFrame, default_power_threshold : float = 1.0, default_power_range : float = 2.0, default_power_draw : float = 30, fault_time : int = 3):
        alarm_tag = 'BLWNFSE'
        type_default_dict = {'PowerIn' : default_power_draw}
        self.default_power_threshold = default_power_threshold
        self.default_power_range = default_power_range
        self.fault_time = fault_time
        super().__init__(bounds_df, alarm_tag,type_default_dict, two_part_tag = False)

    def specific_alarm_function(self, df: pd.DataFrame, daily_df : pd.DataFrame, config : ConfigManager):
        for var_name in self.bounds_df['variable_name'].unique():
            for day in daily_df.index:
                next_day = day + pd.Timedelta(days=1)
                filtered_df = df.loc[(df.index >= day) & (df.index < next_day)]
                rows = self.bounds_df[self.bounds_df['variable_name'] == var_name]
                expected_power_draw = rows.iloc[0]['bound']
                if len(rows) != 1:
                    raise Exception(f"Multiple blown fuse alarm codes for {var_name}")
                if var_name in filtered_df.columns:
                    # Check for consecutive minutes where both power and temp exceed thresholds
                    power_on_mask = filtered_df[var_name] > self.default_power_threshold
                    unexpected_power_mask = filtered_df[var_name] < expected_power_draw - self.default_power_range
                    combined_mask = power_on_mask & unexpected_power_mask

                    # Check for fault_time consecutive minutes
                    consecutive_condition = combined_mask.rolling(window=self.fault_time).min() == 1
                    if consecutive_condition.any():

                         # Find all streaks of consecutive True values
                        group = (consecutive_condition != consecutive_condition.shift()).cumsum()

                        # Iterate through each streak and add an alarm for each
                        for group_id in consecutive_condition.groupby(group).first()[lambda x: x].index:
                            streak_indices = consecutive_condition[group == group_id].index
                            start_time = streak_indices[0] - pd.Timedelta(minutes=self.fault_time-1)
                            end_time = streak_indices[-1]

                            self._add_an_alarm(start_time, end_time, var_name,
                                f"Blown Fuse: {var_name} had a power draw less than {expected_power_draw - self.default_power_range:.1f} while element was ON starting at {start_time}.",
                                certainty="high")
                            
                        # first_true_index = consecutive_condition.idxmax()
                        # adjusted_time = first_true_index - pd.Timedelta(minutes=self.fault_time-1)
                        # _add_an_alarm(alarms, day, var_name, f"Blown Fuse: {var_name} had a power draw less than {expected_power_draw - self.default_power_range:.1f} while element was ON starting at {adjusted_time}.")



from .event_tracking import *
from .Alarm import Alarm
from .alarms.ShortCycle import ShortCycle
from .alarms.TempRange import TempRange
from .alarms.LSInconsist import LSInconsist
from .alarms.SOOChange import SOOChange
from .alarms.BlownFuse import BlownFuse
from .alarms.HPWHOutage import HPWHOutage
from .alarms.BackupUse import BackupUse
from .alarms.HPWHOutlet import HPWHOutlet
from .alarms.HPWHInlet import HPWHInlet
from .alarms.BalancingValve import BalancingValve
from .alarms.TMSetpoint import TMSetpoint
from .alarms.AbnormalCOP import AbnormalCOP
from .alarms.PowerRatio import PowerRatio
from .alarms.Boundary import Boundary

__all__ = ['central_alarm_df_creator','flag_boundary_alarms','power_ratio_alarm','flag_abnormal_COP','flag_high_tm_setpoint',
           'flag_recirc_balance_valve','flag_hp_inlet_temp','flag_backup_use','flag_blown_fuse','flag_unexpected_soo_change','flag_shortcycle',
           'flag_hp_outlet_temp','flag_HP_outage','flag_unexpected_temp','flag_ls_mode_inconsistancy','Alarm','ShortCycle','TempRange','LSInconsist',
           'SOOChange','BlownFuse','HPWHOutage','BackupUse','HPWHOutlet','HPWHInlet','BalancingValve','TMSetpoint','AbnormalCOP','PowerRatio','Boundary']


import pandas as pd
import numpy as np
import datetime as datetime
from ecopipeline import ConfigManager
import re
import mysql.connector.errors as mysqlerrors
from datetime import timedelta
from .alarms.ShortCycle import ShortCycle
from .alarms.TempRange import TempRange
from .alarms.LSInconsist import LSInconsist
from .alarms.SOOChange import SOOChange
from .alarms.BlownFuse import BlownFuse
from .alarms.HPWHOutage import HPWHOutage
from .alarms.BackupUse import BackupUse
from .alarms.HPWHOutlet import HPWHOutlet
from .alarms.HPWHInlet import HPWHInlet
from .alarms.BalancingValve import BalancingValve
from .alarms.TMSetpoint import TMSetpoint
from .alarms.AbnormalCOP import AbnormalCOP
from .alarms.PowerRatio import PowerRatio
from .alarms.Boundary import Boundary

def central_alarm_df_creator(df: pd.DataFrame, daily_data : pd.DataFrame, config : ConfigManager, system: str = "", 
                             default_cop_high_bound : float = 4.5, default_cop_low_bound : float = 0,
                             default_boundary_fault_time : int = 15, site_name : str = None, day_table_name_header : str = "day",
                             power_ratio_period_days : int = 7) -> pd.DataFrame:
    if df.empty:
        print("cannot flag missing balancing valve alarms. Dataframe is empty")
        return pd.DataFrame()
    variable_names_path = config.get_var_names_path()
    try:
        bounds_df = pd.read_csv(variable_names_path)
    except FileNotFoundError:
        print("File Not Found: ", variable_names_path)
        return pd.DataFrame()
    if (system != ""):
        if not 'system' in bounds_df.columns:
            raise Exception("system parameter is non null, however, system is not present in Variable_Names.csv")
        bounds_df = bounds_df.loc[bounds_df['system'] == system]
    
    day_list = daily_data.index.to_list()
    print('Checking for alarms...')
    alarm_df = _convert_silent_alarm_dict_to_df({})
    dict_of_alarms = {}
    dict_of_alarms['boundary'] = Boundary(bounds_df, default_fault_time= default_boundary_fault_time)
    # flag_boundary_alarms(df, config, full_days=day_list, system=system, default_fault_time= default_boundary_fault_time)
    dict_of_alarms['power ratio'] = PowerRatio(bounds_df, day_table_name = config.get_table_name(day_table_name_header), ratio_period_days=power_ratio_period_days)
    # power_ratio_alarm(daily_data, config, day_table_name = config.get_table_name(day_table_name_header), system=system, ratio_period_days=power_ratio_period_days)
    dict_of_alarms['abnormal COP'] = AbnormalCOP(bounds_df, default_high_bound=default_cop_high_bound, default_low_bound=default_cop_low_bound)
    # flag_abnormal_COP(daily_data, config, system = system, default_high_bound=default_cop_high_bound, default_low_bound=default_cop_low_bound)
    dict_of_alarms['temperature maintenance setpoint'] = TMSetpoint(bounds_df)
    # flag_high_tm_setpoint(df, daily_data, config, system=system)
    dict_of_alarms['recirculation loop balancing valve'] = BalancingValve(bounds_df)
    # flag_recirc_balance_valve(daily_data, config, system=system)
    dict_of_alarms['HPWH inlet temperature'] = HPWHInlet(bounds_df)
    # flag_hp_inlet_temp(df, daily_data, config, system)
    dict_of_alarms['HPWH outlet temperature'] = HPWHOutlet(bounds_df)
    # flag_hp_outlet_temp(df, daily_data, config, system)
    dict_of_alarms['improper backup heating use'] = BackupUse(bounds_df)
    # flag_backup_use(df, daily_data, config, system)
    dict_of_alarms['HPWH outage'] = HPWHOutage(bounds_df, day_table_name = config.get_table_name(day_table_name_header))
    # flag_HP_outage(df, daily_data, config, day_table_name = config.get_table_name(day_table_name_header), system=system)
    dict_of_alarms['blown equipment fuse'] = BlownFuse(bounds_df)
    # flag_blown_fuse(df, daily_data, config, system)
    dict_of_alarms['unexpected SOO change'] = SOOChange(bounds_df)
    # flag_unexpected_soo_change(df, daily_data, config, system)
    dict_of_alarms['short cycle'] = ShortCycle(bounds_df)
    # flag_shortcycle(df, daily_data, config, system)
    dict_of_alarms['unexpected temperature'] = TempRange(bounds_df)
    # flag_unexpected_temp(df, daily_data, config, system)
    dict_of_alarms['demand response inconsistency'] = LSInconsist(bounds_df)
    # flag_ls_mode_inconsistancy(df, daily_data, config, system)
    # return alarm.find_alarms(df, daily_df, config)

    ongoing_COP_exception = ['abnormal COP']
    for key, value in dict_of_alarms.items():
        # if key in ongoing_COP_exception and _check_if_during_ongoing_cop_alarm(daily_data, config, site_name):
        #     print("Ongoing DATA_LOSS_COP detected. ABNORMAL_COP events will be uploaded")
        specific_alarm_df = value.find_alarms(df, daily_data, config)
        if len(specific_alarm_df) > 0:
            print(f"Detected {key} alarm(s). Adding to event df...")
            alarm_df = pd.concat([alarm_df, specific_alarm_df])
        else:
            print(f"No {key} alarm(s) detected.")

    # for key, value in dict_of_alarms.items():
    #     if key in ongoing_COP_exception and _check_if_during_ongoing_cop_alarm(daily_data, config, site_name):
    #         print("Ongoing DATA_LOSS_COP detected. No further DATA_LOSS_COP events will be uploaded")
    #     elif len(value) > 0:
    #         print(f"Detected {key} alarm(s). Adding to event df...")
    #         alarm_df = pd.concat([alarm_df, value])
    #     else:
    #         print(f"No {key} alarm(s) detected.")

    return alarm_df

def flag_abnormal_COP(daily_data: pd.DataFrame, config : ConfigManager, system: str = "", default_high_bound : float = 4.5, default_low_bound : float = 0) -> pd.DataFrame:
    variable_names_path = config.get_var_names_path()
    try:
        bounds_df = pd.read_csv(variable_names_path)
    except FileNotFoundError:
        print("File Not Found: ", variable_names_path)
        return pd.DataFrame()
    alarm = AbnormalCOP(bounds_df, default_high_bound, default_low_bound)
    return alarm.find_alarms(None, daily_data, config)


def flag_boundary_alarms(df: pd.DataFrame, config : ConfigManager, default_fault_time : int = 15, system: str = "", full_days : list = None) -> pd.DataFrame:
    """
    Function will take a pandas dataframe and location of alarm information in a csv,
    and create an dataframe with applicable alarm events

    Parameters
    ----------
    df: pd.DataFrame
        post-transformed dataframe for minute data. It should be noted that this function expects consecutive, in order minutes. If minutes
        are out of order or have gaps, the function may return erroneous alarms.
    config : ecopipeline.ConfigManager
        The ConfigManager object that holds configuration data for the pipeline. Among other things, this object will point to a file 
        called Variable_Names.csv in the input folder of the pipeline (e.g. "full/path/to/pipeline/input/Variable_Names.csv").
        The file must have at least three columns which must be titled "variable_name", "low_alarm", and "high_alarm" which should contain the
        name of each variable in the dataframe that requires the alarming, the lower bound for acceptable data, and the upper bound for
        acceptable data respectively
    default_fault_time : int
        Number of consecutive minutes that a sensor must be out of bounds for to trigger an alarm. Can be customized for each variable with 
        the fault_time column in Variable_Names.csv
    system: str
        string of system name if processing a particular system in a Variable_Names.csv file with multiple systems. Leave as an empty string if not aplicable.
    full_days : list
        list of pd.Datetimes that should be considered full days here. If set to none, will take any day at all present in df

    Returns
    ------- 
    pd.DataFrame:
        Pandas dataframe with alarm events
    """
    if df.empty:
        print("cannot flag boundary alarms. Dataframe is empty")
        return pd.DataFrame()
    variable_names_path = config.get_var_names_path()
    try:
        bounds_df = pd.read_csv(variable_names_path)
    except FileNotFoundError:
        print("File Not Found: ", variable_names_path)
        return pd.DataFrame()
    alarm = Boundary(bounds_df, default_fault_time)
    return alarm.find_alarms(df, None, config)

def flag_high_tm_setpoint(df: pd.DataFrame, daily_df: pd.DataFrame, config : ConfigManager, default_fault_time : int = 3, 
                             system: str = "", default_setpoint : float = 130.0, default_power_indication : float = 1.0,
                             default_power_ratio : float = 0.4) -> pd.DataFrame:
    """
    Function will take a pandas dataframe and location of alarm information in a csv,
    and create an dataframe with applicable alarm events

    VarNames syntax:
    TMSTPT_T_ID:### - Swing Tank Outlet Temperature. Alarm triggered if over number ### (or 130) for 3 minutes with power on
    TMSTPT_SP_ID:### - Swing Tank Power. ### is lowest recorded power for Swing Tank to be considered 'on'. Defaults to 1.0
    TMSTPT_TP_ID:### - Total System Power for ratio alarming for alarming if swing tank power is more than ### (40% default) of usage
    TMSTPT_ST_ID:### - Swing Tank Setpoint that should not change at all from ### (default 130)

    Parameters
    ----------
    df: pd.DataFrame
        post-transformed dataframe for minute data. It should be noted that this function expects consecutive, in order minutes. If minutes
        are out of order or have gaps, the function may return erroneous alarms.
    daily_df: pd.DataFrame
        post-transformed dataframe for daily data. Used for checking power ratios and determining which days to process.
    config : ecopipeline.ConfigManager
        The ConfigManager object that holds configuration data for the pipeline. Among other things, this object will point to a file
        called Variable_Names.csv in the input folder of the pipeline (e.g. "full/path/to/pipeline/input/Variable_Names.csv").
        The file must have at least two columns which must be titled "variable_name" and "alarm_codes" which should contain the
        name of each variable in the dataframe that requires alarming and the TMSTPT alarm codes (e.g., TMSTPT_T_1:140, TMSTPT_SP_1:2.0)
    default_fault_time : int
        Number of consecutive minutes for T+SP alarms (default 3). T+SP alarms trigger when tank is powered and temperature exceeds
        setpoint for this many consecutive minutes.
    system: str
        string of system name if processing a particular system in a Variable_Names.csv file with multiple systems. Leave as an empty string if not aplicable.
    default_setpoint : float
        Default temperature setpoint in degrees for T and ST alarm codes when no custom bound is specified (default 130.0)
    default_power_indication : float
        Default power threshold in kW for SP alarm codes when no custom bound is specified (default 1.0)
    default_power_ratio : float
        Default power ratio threshold (as decimal, e.g., 0.4 for 40%) for TP alarm codes when no custom bound is specified (default 0.4)

    Returns
    ------- 
    pd.DataFrame:
        Pandas dataframe with alarm events
    """
    if df.empty:
        print("cannot flag swing tank setpoint alarms. Dataframe is empty")
        return pd.DataFrame()
    variable_names_path = config.get_var_names_path()
    try:
        bounds_df = pd.read_csv(variable_names_path)
    except FileNotFoundError:
        print("File Not Found: ", variable_names_path)
        return pd.DataFrame()
    alarm = TMSetpoint(bounds_df, default_fault_time, default_setpoint, default_power_indication, default_power_ratio)
    return alarm.find_alarms(df, daily_df, config)

def flag_backup_use(df: pd.DataFrame, daily_df: pd.DataFrame, config : ConfigManager, 
                             system: str = "", default_setpoint : float = 130.0, default_power_ratio : float = 0.1) -> pd.DataFrame:
    """
    Function will take a pandas dataframe and location of alarm information in a csv,
    and create an dataframe with applicable alarm events

    VarNames syntax:
    BU_P_ID - Back Up Tank Power Variable. Must be in same power units as total system power
    BU_TP_ID:### - Total System Power for ratio alarming for alarming if back up power is more than ### (40% default) of usage
    BU_ST_ID:### - Back Up Setpoint that should not change at all from ### (default 130)

    Parameters
    ----------
    df: pd.DataFrame
        post-transformed dataframe for minute data. It should be noted that this function expects consecutive, in order minutes. If minutes
        are out of order or have gaps, the function may return erroneous alarms.
    daily_df: pd.DataFrame
        post-transformed dataframe for daily data. Used for checking power ratios and determining which days to process.
    config : ecopipeline.ConfigManager
        The ConfigManager object that holds configuration data for the pipeline. Among other things, this object will point to a file
        called Variable_Names.csv in the input folder of the pipeline (e.g. "full/path/to/pipeline/input/Variable_Names.csv").
        The file must have at least two columns which must be titled "variable_name" and "alarm_codes" which should contain the
        name of each variable in the dataframe that requires alarming and the STS alarm codes (e.g., STS_T_1:140, STS_SP_1:2.0)
    system: str
        string of system name if processing a particular system in a Variable_Names.csv file with multiple systems. Leave as an empty string if not aplicable.
    default_setpoint : float
        Default temperature setpoint in degrees for T and ST alarm codes when no custom bound is specified (default 130.0)
    default_power_indication : float
        Default power threshold in kW for SP alarm codes when no custom bound is specified (default 1.0)
    default_power_ratio : float
        Default power ratio threshold (as decimal, e.g., 0.4 for 40%) for TP alarm codes when no custom bound is specified (default 0.4)

    Returns
    ------- 
    pd.DataFrame:
        Pandas dataframe with alarm events
    """
    if df.empty:
        print("cannot flag swing tank setpoint alarms. Dataframe is empty")
        return pd.DataFrame()
    variable_names_path = config.get_var_names_path()
    try:
        bounds_df = pd.read_csv(variable_names_path)
    except FileNotFoundError:
        print("File Not Found: ", variable_names_path)
        return pd.DataFrame()
    alarm = BackupUse(bounds_df,  default_setpoint, default_power_ratio)
    return alarm.find_alarms(df, daily_df, config)

def flag_HP_outage(df: pd.DataFrame, daily_df: pd.DataFrame, config : ConfigManager, day_table_name : str, system: str = "", default_power_ratio : float = 0.3,
                   ratio_period_days : int = 7) -> pd.DataFrame:
    """
    Detects possible heat pump failures or outages by checking if heat pump power consumption falls below
    an expected ratio of total system power over a rolling period, or by checking for non-zero values in
    a direct alarm variable from the heat pump controller.

    VarNames syntax:
    HPOUT_POW_[OPTIONAL ID]:### - Heat pump power variable. ### is the minimum expected ratio of HP power to total power
        (default 0.3 for 30%). Must be in same power units as total system power.
    HPOUT_TP_[OPTIONAL ID] - Total system power variable for ratio comparison. Required when using POW codes.
    HPOUT_ALRM_[OPTIONAL ID] - Direct alarm variable from HP controller. Alarm triggers if any non-zero value is detected.

    Parameters
    ----------
    df: pd.DataFrame
        Post-transformed dataframe for minute data. Used for checking ALRM codes for non-zero values.
    daily_df: pd.DataFrame
        Post-transformed dataframe for daily data. Used for checking power ratios over the rolling period.
    config : ecopipeline.ConfigManager
        The ConfigManager object that holds configuration data for the pipeline. Among other things, this object will point to a file
        called Variable_Names.csv in the input folder of the pipeline (e.g. "full/path/to/pipeline/input/Variable_Names.csv").
        The file must have at least two columns which must be titled "variable_name" and "alarm_codes" which should contain the
        name of each variable in the dataframe that requires alarming and the HPOUT alarm codes (e.g., HPOUT_POW_1:0.3, HPOUT_TP_1, HPOUT_ALRM_1).
    day_table_name : str
        Name of the daily database table to fetch previous days' data for the rolling period calculation.
    system: str
        String of system name if processing a particular system in a Variable_Names.csv file with multiple systems. Leave as an empty string if not applicable.
    default_power_ratio : float
        Default minimum power ratio threshold (as decimal, e.g., 0.3 for 30%) for POW alarm codes when no custom bound is specified (default 0.3).
        An alarm triggers if HP power falls below this ratio of total power over the rolling period.
    ratio_period_days : int
        Number of days to use for the rolling power ratio calculation (default 7). Must be greater than 1.

    Returns
    -------
    pd.DataFrame:
        Pandas dataframe with alarm events
    """
    if df.empty:
        print("cannot flag swing tank setpoint alarms. Dataframe is empty")
        return pd.DataFrame()
    variable_names_path = config.get_var_names_path()
    try:
        bounds_df = pd.read_csv(variable_names_path)
    except FileNotFoundError:
        print("File Not Found: ", variable_names_path)
        return pd.DataFrame()
    
    alarm = HPWHOutage(bounds_df,  day_table_name, default_power_ratio, ratio_period_days)
    return alarm.find_alarms(df, daily_df, config) 

def flag_recirc_balance_valve(daily_df: pd.DataFrame, config : ConfigManager, system: str = "", default_power_ratio : float = 0.4) -> pd.DataFrame:
    """
    Detects recirculation balance issues by comparing sum of ER (equipment recirculation) heater
    power to either total power or heating output.

    VarNames syntax:
    BV_ER_[OPTIONAL ID] - Indicates a power variable for an ER heater (equipment recirculation).
        Multiple ER variables with the same ID will be summed together.
    BV_TP_[OPTIONAL ID]:### - Indicates the Total Power of the system. Optional ### for the percentage
        threshold that should not be crossed by the ER elements (default 0.4 for 40%).
        Alarm triggers when sum of ER >= total_power * threshold.
    BV_OUT_[OPTIONAL ID] - Indicates the heating output variable the ER heating contributes to.
        Alarm triggers when sum of ER > sum of OUT * 0.95 (i.e., ER exceeds 95% of heating output).
        Multiple OUT variables with the same ID will be summed together.

    Note: Each alarm ID requires at least one ER code AND either one TP code OR at least one OUT code.
    If a TP code exists for an ID, it takes precedence over OUT codes.

    Parameters
    ----------
    daily_df: pd.DataFrame
        Post-transformed dataframe for daily data. Used for checking recirculation balance by comparing sum of ER equipment
        power to total power or heating output power.
    config : ecopipeline.ConfigManager
        The ConfigManager object that holds configuration data for the pipeline. Among other things, this object will point to a file
        called Variable_Names.csv in the input folder of the pipeline (e.g. "full/path/to/pipeline/input/Variable_Names.csv").
        The file must have at least two columns which must be titled "variable_name" and "alarm_codes" which should contain the
        name of each variable in the dataframe that requires alarming and the BV alarm codes (e.g., BV_ER_1, BV_TP_1:0.3)
    system: str
        String of system name if processing a particular system in a Variable_Names.csv file with multiple systems. Leave as an empty string if not applicable.
    default_power_ratio : float
        Default power ratio threshold (as decimal, e.g., 0.4 for 40%) for TP alarm codes when no custom bound is specified (default 0.4).

    Returns
    -------
    pd.DataFrame:
        Pandas dataframe with alarm events
    """
    if daily_df.empty:
        print("cannot flag missing balancing valve alarms. Dataframe is empty")
        return pd.DataFrame()
    variable_names_path = config.get_var_names_path()
    try:
        bounds_df = pd.read_csv(variable_names_path)
    except FileNotFoundError:
        print("File Not Found: ", variable_names_path)
        return pd.DataFrame()
    alarm = BalancingValve(bounds_df, default_power_ratio)
    return alarm.find_alarms(None, daily_df, config)

def flag_hp_inlet_temp(df: pd.DataFrame, daily_df: pd.DataFrame, config : ConfigManager, system: str = "", default_power_threshold : float = 1.0,
                       default_temp_threshold : float = 115.0, fault_time : int = 5) -> pd.DataFrame:
    """
    Function will take a pandas dataframe and location of alarm information in a csv,
    and create an dataframe with applicable alarm events

    VarNames syntax:
    HPI_POW_[OPTIONAL ID]:### - Indicates a power variable for the heat pump. ### is the power threshold (default 1.0) above which
        the heat pump is considered 'on'
    HPI_T_[OPTIONAL ID]:### - Indicates heat pump inlet temperature variable. ### is the temperature threshold (default 120.0)
        that should not be exceeded while the heat pump is on

    Parameters
    ----------
    df: pd.DataFrame
        post-transformed dataframe for minute data. It should be noted that this function expects consecutive, in order minutes. If minutes
        are out of order or have gaps, the function may return erroneous alarms.
    daily_df: pd.DataFrame
        post-transformed dataframe for daily data.
    config : ecopipeline.ConfigManager
        The ConfigManager object that holds configuration data for the pipeline. Among other things, this object will point to a file
        called Variable_Names.csv in the input folder of the pipeline (e.g. "full/path/to/pipeline/input/Variable_Names.csv").
        The file must have at least two columns which must be titled "variable_name" and "alarm_codes" which should contain the
        name of each variable in the dataframe that requires alarming and the HPI alarm codes (e.g., HPI_POW_1:0.5, HPI_T_1:125.0)
    system: str
        string of system name if processing a particular system in a Variable_Names.csv file with multiple systems. Leave as an empty string if not aplicable.
    default_power_threshold : float
        Default power threshold for POW alarm codes when no custom bound is specified (default 0.4). Heat pump is considered 'on'
        when power exceeds this value.
    default_temp_threshold : float
        Default temperature threshold for T alarm codes when no custom bound is specified (default 120.0). Alarm triggers when
        temperature exceeds this value while heat pump is on.
    fault_time : int
        Number of consecutive minutes that both power and temperature must exceed their thresholds before triggering an alarm (default 10).

    Returns
    -------
    pd.DataFrame:
        Pandas dataframe with alarm events
    """
    if df.empty:
        print("cannot flag missing balancing valve alarms. Dataframe is empty")
        return pd.DataFrame()
    variable_names_path = config.get_var_names_path()
    try:
        bounds_df = pd.read_csv(variable_names_path)
    except FileNotFoundError:
        print("File Not Found: ", variable_names_path)
        return pd.DataFrame()
    alarm = HPWHInlet(bounds_df, default_power_threshold, default_temp_threshold, fault_time)
    return alarm.find_alarms(df, daily_df, config)

def flag_hp_outlet_temp(df: pd.DataFrame, daily_df: pd.DataFrame, config : ConfigManager, system: str = "", default_power_threshold : float = 1.0,
                       default_temp_threshold : float = 140.0, fault_time : int = 5) -> pd.DataFrame:
    """
    Detects low heat pump outlet temperature by checking if the outlet temperature falls below a threshold
    while the heat pump is running. The first 10 minutes after each HP turn-on are excluded as a warmup
    period. An alarm triggers if the temperature stays below the threshold for `fault_time` consecutive
    minutes after the warmup period.

    VarNames syntax:
    HPO_POW_[OPTIONAL ID]:### - Indicates a power variable for the heat pump. ### is the power threshold (default 1.0) above which
        the heat pump is considered 'on'.
    HPO_T_[OPTIONAL ID]:### - Indicates heat pump outlet temperature variable. ### is the temperature threshold (default 140.0)
        that should always be exceeded while the heat pump is on after the 10-minute warmup period.

    Parameters
    ----------
    df: pd.DataFrame
        Post-transformed dataframe for minute data. It should be noted that this function expects consecutive, in order minutes. If minutes
        are out of order or have gaps, the function may return erroneous alarms.
    daily_df: pd.DataFrame
        Post-transformed dataframe for daily data.
    config : ecopipeline.ConfigManager
        The ConfigManager object that holds configuration data for the pipeline. Among other things, this object will point to a file
        called Variable_Names.csv in the input folder of the pipeline (e.g. "full/path/to/pipeline/input/Variable_Names.csv").
        The file must have at least two columns which must be titled "variable_name" and "alarm_codes" which should contain the
        name of each variable in the dataframe that requires alarming and the HPO alarm codes (e.g., HPO_POW_1:1.0, HPO_T_1:140.0).
    system: str
        String of system name if processing a particular system in a Variable_Names.csv file with multiple systems. Leave as an empty string if not applicable.
    default_power_threshold : float
        Default power threshold for POW alarm codes when no custom bound is specified (default 1.0). Heat pump is considered 'on'
        when power exceeds this value.
    default_temp_threshold : float
        Default temperature threshold for T alarm codes when no custom bound is specified (default 140.0). Alarm triggers when
        temperature falls BELOW this value while heat pump is on (after warmup period).
    fault_time : int
        Number of consecutive minutes that temperature must be below threshold (after warmup) before triggering an alarm (default 5).

    Returns
    -------
    pd.DataFrame:
        Pandas dataframe with alarm events
    """
    if df.empty:
        print("cannot flag missing balancing valve alarms. Dataframe is empty")
        return pd.DataFrame()
    variable_names_path = config.get_var_names_path()
    try:
        bounds_df = pd.read_csv(variable_names_path)
    except FileNotFoundError:
        print("File Not Found: ", variable_names_path)
        return pd.DataFrame()
    
    alarm = HPWHOutlet(bounds_df, default_power_threshold, default_temp_threshold, fault_time)
    return alarm.find_alarms(df, daily_df, config)

def flag_blown_fuse(df: pd.DataFrame, daily_df: pd.DataFrame, config : ConfigManager, system: str = "", default_power_threshold : float = 1.0,
                       default_power_range : float = 2.0, default_power_draw : float = 30, fault_time : int = 3) -> pd.DataFrame:
    """
    Detects blown fuse alarms for heating elements by identifying when an element is drawing power
    but significantly less than expected, which may indicate a blown fuse.

    VarNames syntax:
    BF_[OPTIONAL ID]:### - Indicates a blown fuse alarm for an element. ### is the expected kW input when the element is on.

    Parameters
    ----------
    df: pd.DataFrame
        Post-transformed dataframe for minute data. It should be noted that this function expects consecutive, in order minutes. If minutes
        are out of order or have gaps, the function may return erroneous alarms.
    daily_df: pd.DataFrame
        Post-transformed dataframe for daily data.
    config : ecopipeline.ConfigManager
        The ConfigManager object that holds configuration data for the pipeline. Among other things, this object will point to a file
        called Variable_Names.csv in the input folder of the pipeline (e.g. "full/path/to/pipeline/input/Variable_Names.csv").
        The file must have at least two columns which must be titled "variable_name" and "alarm_codes" which should contain the
        name of each variable in the dataframe that requires alarming and the BF alarm codes (e.g., BF:30, BF_1:25).
    system: str
        String of system name if processing a particular system in a Variable_Names.csv file with multiple systems. Leave as an empty string if not applicable.
    default_power_threshold : float
        Power threshold to determine if the element is "on" (default 1.0). Element is considered on when power exceeds this value.
    default_power_range : float
        Allowable variance below the expected power draw (default 2.0). An alarm triggers when the actual power draw is less than
        (expected_power_draw - default_power_range) while the element is on.
    default_power_draw : float
        Default expected power draw in kW when no custom bound is specified in the alarm code (default 30).
    fault_time : int
        Number of consecutive minutes that the fault condition must persist before triggering an alarm (default 3).

    Returns
    -------
    pd.DataFrame:
        Pandas dataframe with alarm events
    """
    if df.empty:
        print("cannot flag missing balancing valve alarms. Dataframe is empty")
        return pd.DataFrame()
    variable_names_path = config.get_var_names_path()
    try:
        bounds_df = pd.read_csv(variable_names_path)
    except FileNotFoundError:
        print("File Not Found: ", variable_names_path)
        return pd.DataFrame()

    alarm = BlownFuse(bounds_df, default_power_threshold, default_power_range, default_power_draw,fault_time)
    return alarm.find_alarms(df, daily_df, config)

def flag_unexpected_soo_change(df: pd.DataFrame, daily_df: pd.DataFrame, config : ConfigManager, system: str = "", default_power_threshold : float = 1.0,
                       default_on_temp : float = 115.0, default_off_temp : float = 140.0) -> pd.DataFrame:
    """
    Detects unexpected state of operation (SOO) changes by checking if the heat pump turns on or off
    when the temperature is not near the expected aquastat setpoint thresholds. An alarm is triggered
    if the HP turns on/off and the corresponding temperature is more than 5.0 degrees away from the
    expected threshold.

    VarNames syntax:
    SOOCHNG_POW:### - Indicates a power variable for the heat pump system (should be total power across all primary heat pumps). ### is the power threshold (default 1.0) above which
        the heat pump system is considered 'on'.
    SOOCHNG_ON_[Mode ID]:### - Indicates the temperature variable at the ON aquastat fraction. ### is the temperature (default 115.0)
        that should trigger the heat pump to turn ON. Mode ID should be the load up mode from ['loadUp','shed','criticalPeak','gridEmergency','advLoadUp','normal'] or left blank for normal mode
    SOOCHNG_OFF_[Mode ID]:### - Indicates the temperature variable at the OFF aquastat fraction (can be same as ON aquastat). ### is the temperature (default 140.0)
        that should trigger the heat pump to turn OFF. Mode ID should be the load up mode from ['loadUp','shed','criticalPeak','gridEmergency','advLoadUp','normal'] or left blank for normal mode

    Parameters
    ----------
    df: pd.DataFrame
        Post-transformed dataframe for minute data. It should be noted that this function expects consecutive, in order minutes. If minutes
        are out of order or have gaps, the function may return erroneous alarms.
    daily_df: pd.DataFrame
        Post-transformed dataframe for daily data.
    config : ecopipeline.ConfigManager
        The ConfigManager object that holds configuration data for the pipeline. Among other things, this object will point to a file
        called Variable_Names.csv in the input folder of the pipeline (e.g. "full/path/to/pipeline/input/Variable_Names.csv").
        The file must have at least two columns which must be titled "variable_name" and "alarm_codes" which should contain the
        name of each variable in the dataframe that requires alarming and the SOOCHNG alarm codes (e.g., SOOCHNG_POW_normal:1.0, SOOCHNG_ON_normal:115.0, SOOCHNG_OFF_normal:140.0).
    system: str
        String of system name if processing a particular system in a Variable_Names.csv file with multiple systems. Leave as an empty string if not applicable.
    default_power_threshold : float
        Default power threshold for POW alarm codes when no custom bound is specified (default 1.0). Heat pump is considered 'on'
        when power exceeds this value.
    default_on_temp : float
        Default ON temperature threshold (default 115.0). When the HP turns on, an alarm triggers if the temperature
        is more than 5.0 degrees away from this value.
    default_off_temp : float
        Default OFF temperature threshold (default 140.0). When the HP turns off, an alarm triggers if the temperature
        is more than 5.0 degrees away from this value.

    Returns
    -------
    pd.DataFrame:
        Pandas dataframe with alarm events
    """
    if df.empty:
        print("cannot flag missing balancing valve alarms. Dataframe is empty")
        return pd.DataFrame()
    variable_names_path = config.get_var_names_path()
    try:
        bounds_df = pd.read_csv(variable_names_path)
    except FileNotFoundError:
        print("File Not Found: ", variable_names_path)
        return pd.DataFrame()

    alarm = SOOChange(bounds_df, default_power_threshold, default_on_temp, default_off_temp)
    return alarm.find_alarms(df, daily_df, config)

def flag_ls_mode_inconsistancy(df: pd.DataFrame, daily_df: pd.DataFrame, config : ConfigManager, system: str = "") -> pd.DataFrame:
    """
    Detects when reported loadshift mode does not match its expected value during a load shifting event.
    An alarm is triggered if the variable value does not equal the expected value during the
    time periods defined in the load shifting schedule for that mode.

    VarNames syntax:
    SOO_[mode]:### - Indicates a variable that should equal ### during [mode] load shifting events.
        [mode] can be: normal, loadUp, shed, criticalPeak, gridEmergency, advLoadUp
        ### is the expected value (e.g., SOO_loadUp:1 means the variable should be 1 during loadUp events)

    Parameters
    ----------
    df: pd.DataFrame
        Post-transformed dataframe for minute data. It should be noted that this function expects consecutive,
        in order minutes. If minutes are out of order or have gaps, the function may return erroneous alarms.
    daily_df: pd.DataFrame
        Pandas dataframe with daily data. This dataframe should have a datetime index.
    config : ecopipeline.ConfigManager
        The ConfigManager object that holds configuration data for the pipeline.
    system: str
        String of system name if processing a particular system in a Variable_Names.csv file with multiple systems.

    Returns
    -------
    pd.DataFrame:
        Pandas dataframe with alarm events
    """
    if df.empty:
        print("cannot flag load shift mode inconsistency alarms. Dataframe is empty")
        return pd.DataFrame()
    variable_names_path = config.get_var_names_path()
    try:
        bounds_df = pd.read_csv(variable_names_path)
    except FileNotFoundError:
        print("File Not Found: ", variable_names_path)
        return pd.DataFrame()
    
    alarm = LSInconsist(bounds_df)
    return alarm.find_alarms(df, daily_df, config)

def flag_unexpected_temp(df: pd.DataFrame, daily_df: pd.DataFrame, config : ConfigManager, system: str = "", default_high_temp : float = 130,
                       default_low_temp : float = 115, fault_time : int = 10) -> pd.DataFrame:
    """
    Detects when a temperature value falls outside an acceptable range for
    too long. An alarm is triggered if the temperature is above the high bound or below the low bound
    for `fault_time` consecutive minutes.

    VarNames syntax:
    TMPRNG_[OPTIONAL ID]:###-### - Indicates a temperature variable. ###-### is the acceptable temperature range
        (e.g., TMPRNG:110-130 means temperature should stay between 110 and 130 degrees).

    Parameters
    ----------
    df: pd.DataFrame
        Post-transformed dataframe for minute data. It should be noted that this function expects consecutive, in order minutes. If minutes
        are out of order or have gaps, the function may return erroneous alarms.
    daily_df: pd.DataFrame
        Post-transformed dataframe for daily data. Used for determining which days to process.
    config : ecopipeline.ConfigManager
        The ConfigManager object that holds configuration data for the pipeline. Among other things, this object will point to a file
        called Variable_Names.csv in the input folder of the pipeline (e.g. "full/path/to/pipeline/input/Variable_Names.csv").
        The file must have at least two columns which must be titled "variable_name" and "alarm_codes" which should contain the
        name of each variable in the dataframe that requires alarming and the DHW alarm codes (e.g., DHW:110-130, DHW_1:115-125).
    system: str
        String of system name if processing a particular system in a Variable_Names.csv file with multiple systems. Leave as an empty string if not applicable.
    default_high_temp : float
        Default high temperature bound when no custom range is specified in the alarm code (default 130). Temperature above this triggers alarm.
    default_low_temp : float
        Default low temperature bound when no custom range is specified in the alarm code (default 130). Temperature below this triggers alarm.
    fault_time : int
        Number of consecutive minutes that temperature must be outside the acceptable range before triggering an alarm (default 10).

    Returns
    -------
    pd.DataFrame:
        Pandas dataframe with alarm events
    """
    if df.empty:
        print("cannot flag missing balancing valve alarms. Dataframe is empty")
        return pd.DataFrame()
    variable_names_path = config.get_var_names_path()
    try:
        bounds_df = pd.read_csv(variable_names_path)
    except FileNotFoundError:
        print("File Not Found: ", variable_names_path)
        return pd.DataFrame()
    temp_alarm = TempRange(bounds_df, default_high_temp, default_low_temp, fault_time)
    return temp_alarm.find_alarms(df, daily_df, config)

def flag_shortcycle(df: pd.DataFrame, daily_df: pd.DataFrame, config : ConfigManager, system: str = "", default_power_threshold : float = 1.0,
                       short_cycle_time : int = 15) -> pd.DataFrame:
    """
    Detects short cycling by identifying when the heat pump turns on for less than `short_cycle_time`
    consecutive minutes before turning off again. Short cycling can indicate equipment issues or
    improper system sizing.

    VarNames syntax:
    SHRTCYC_[OPTIONAL ID]:### - Indicates a power variable for the heat pump. ### is the power threshold (default 1.0) above which
        the heat pump is considered 'on'.

    Parameters
    ----------
    df: pd.DataFrame
        Post-transformed dataframe for minute data. It should be noted that this function expects consecutive, in order minutes. If minutes
        are out of order or have gaps, the function may return erroneous alarms.
    daily_df: pd.DataFrame
        Post-transformed dataframe for daily data.
    config : ecopipeline.ConfigManager
        The ConfigManager object that holds configuration data for the pipeline. Among other things, this object will point to a file
        called Variable_Names.csv in the input folder of the pipeline (e.g. "full/path/to/pipeline/input/Variable_Names.csv").
        The file must have at least two columns which must be titled "variable_name" and "alarm_codes" which should contain the
        name of each variable in the dataframe that requires alarming and the SHRTCYC alarm codes (e.g., SHRTCYC:1.0, SHRTCYC_1:0.5).
    system: str
        String of system name if processing a particular system in a Variable_Names.csv file with multiple systems. Leave as an empty string if not applicable.
    default_power_threshold : float
        Default power threshold when no custom bound is specified in the alarm code (default 1.0). Heat pump is considered 'on'
        when power exceeds this value.
    short_cycle_time : int
        Minimum expected run time in minutes (default 15). An alarm triggers if the heat pump runs for fewer than this many
        consecutive minutes before turning off.

    Returns
    -------
    pd.DataFrame:
        Pandas dataframe with alarm events
    """
    if df.empty:
        print("cannot flag missing balancing valve alarms. Dataframe is empty")
        return pd.DataFrame()
    variable_names_path = config.get_var_names_path()
    try:
        bounds_df = pd.read_csv(variable_names_path)
    except FileNotFoundError:
        print("File Not Found: ", variable_names_path)
        return pd.DataFrame()

    short_alarm = ShortCycle(bounds_df, default_power_threshold, short_cycle_time)
    return short_alarm.find_alarms(df, daily_df, config)


def _convert_silent_alarm_dict_to_df(alarm_dict : dict) -> pd.DataFrame:
    events = {
        'start_time_pt' : [],
        'end_time_pt' : [],
        'alarm_type' : [],
        'alarm_detail' : [],
        'variable_name' : []
    }
    for key, value_list in alarm_dict.items():
        for value in value_list:
            events['start_time_pt'].append(key)
            # Use end_time from value[2] if provided, otherwise use key
            events['end_time_pt'].append(value[2] if len(value) > 2 else key)
            events['alarm_type'].append(value[3] if len(value) > 3 else 'SILENT_ALARM')
            events['alarm_detail'].append(value[1])
            events['variable_name'].append(value[0])

    event_df = pd.DataFrame(events)
    event_df.set_index('start_time_pt', inplace=True)
    return event_df


def power_ratio_alarm(daily_df: pd.DataFrame, config : ConfigManager, day_table_name : str, system: str = "", verbose : bool = False, ratio_period_days : int = 7) -> pd.DataFrame:
    """
    Function will take a pandas dataframe of daily data and location of alarm information in a csv,
    and create an dataframe with applicable alarm events

    Parameters
    ----------
    daily_df: pd.DataFrame
        post-transformed dataframe for daily data. It should be noted that this function expects consecutive, in order days. If days
        are out of order or have gaps, the function may return erroneous alarms.
    config : ecopipeline.ConfigManager
        The ConfigManager object that holds configuration data for the pipeline. Among other things, this object will point to a file 
        called Variable_Names.csv in the input folder of the pipeline (e.g. "full/path/to/pipeline/input/Variable_Names.csv").
        The file must have at least two columns which must be titled "variable_name", "alarm_codes" which should contain the
        name of each variable in the dataframe that requires the alarming and the ratio alarm code in the form "PR_{Power Ratio Name}:{low percentage}-{high percentage}
    system: str
        string of system name if processing a particular system in a Variable_Names.csv file with multiple systems. Leave as an empty string if not aplicable.
    verbose : bool
        add print statements in power ratio

    Returns
    ------- 
    pd.DataFrame:
        Pandas dataframe with alarm events, empty if no alarms triggered
    """
    variable_names_path = config.get_var_names_path()
    try:
        bounds_df = pd.read_csv(variable_names_path)
    except FileNotFoundError:
        print("File Not Found: ", variable_names_path)
        return pd.DataFrame()
    alarm = PowerRatio(bounds_df, day_table_name, ratio_period_days)
    return alarm.find_alarms(None, daily_df, config)



import pandas as pd
import numpy as np
import datetime as datetime
from ecopipeline import ConfigManager
import re
import mysql.connector.errors as mysqlerrors
from datetime import timedelta
from ecopipeline.event_tracking.Alarm import Alarm

class SOOChange(Alarm):
    """
    Detects unexpected state of operation (SOO) changes by checking if the heat pump turns on or off
    when the corresponding temperature is not near the expected aquastat setpoint threshold. An alarm
    triggers if the HP turns on/off and the temperature differs by more than 5.0 degrees from expected.

    Variable_Names.csv configuration:
      The role of each variable is determined by the first underscore-separated part of its variable_name,
      while the alarm_codes column carries the tag, sub-type, and bound:
        variable_name: PowerIn_[name], alarm_codes: SOOCHNG:### — HP system power variable (total power
            across all primary HPs). Bound (###) from alarm_codes is the power threshold (default 1.0).
        variable_name: Temp_[name], alarm_codes: SOOCHNG_ON_[Mode]:### — Temperature variable checked at
            HP turn-on events. Bound (###) from alarm_codes is the expected ON temperature (default 115.0).
        variable_name: Temp_[name], alarm_codes: SOOCHNG_OFF[Mode]:### — Temperature variable checked at
            HP turn-off events. Bound (###) from alarm_codes is the expected OFF temperature (default 140.0).
      [Mode] in alarm_codes is optional and must be one of: loadUp, shed, criticalPeak, gridEmergency, advLoadUp.
        If omitted, the variable applies to normal (non-load-shifting) operation.
      Each mode requires exactly one ON and one OFF temperature variable.

    Parameters
    ----------
    default_power_threshold : float
        Default power threshold for PowerIn variables when no bound is specified (default 1.0).
    default_on_temp : float
        Default ON temperature threshold (default 115.0). Alarm triggers if temperature differs by
        more than 5.0 degrees from this value when the HP turns on.
    default_off_temp : float
        Default OFF temperature threshold (default 140.0). Alarm triggers if temperature differs by
        more than 5.0 degrees from this value when the HP turns off.
    """
    def __init__(self, bounds_df : pd.DataFrame, default_power_threshold : float = 1.0, default_on_temp : float = 115.0, default_off_temp : float = 140.0):
        alarm_tag = 'SOOCHNG'
        self.default_power_threshold = default_power_threshold
        self.default_on_temp = default_on_temp
        self.default_off_temp = default_off_temp
        self.soo_dict = {
                'loadUp' : 'LOAD UP',
                'shed' : 'SHED',
                'criticalPeak': 'CRITICAL PEAK',
                'gridEmergency' : 'GRID EMERGENCY',
                'advLoadUp' : 'ADVANCED LOAD UP'
            }
        type_default_dict = {
                'PowerIn' : default_power_threshold,
                'ON' : default_on_temp,
                'OFF' : default_off_temp
            }
        super().__init__(bounds_df, alarm_tag, type_default_dict)

    def specific_alarm_function(self, df: pd.DataFrame, daily_df : pd.DataFrame, config : ConfigManager):
        ls_df = config.get_ls_df()
        pow_codes = self.bounds_df[self.bounds_df['alarm_code_type'] == 'PowerIn']
        if len(pow_codes) != 1:
            raise Exception(f"Improper alarm codes for SOO changes; must have 1 power variable to indicate power to HPWH(s).")
        pow_var_name = pow_codes.iloc[0]['variable_name']
        pow_thresh = pow_codes.iloc[0]['bound']
        non_pow_bounds_df = self.bounds_df[self.bounds_df['alarm_code_type'] != 'PowerIn']

        for alarm_id in non_pow_bounds_df['alarm_code_id'].unique():
            ls_filtered_df = df.copy()
            soo_mode_name = 'NORMAL'
            if alarm_id in self.soo_dict.keys():
                if not ls_df.empty:
                    # Filter ls_filtered_df for only date ranges in the right mode of ls_df
                    mode_rows = ls_df[ls_df['event'] == alarm_id]
                    mask = pd.Series(False, index=ls_filtered_df.index)
                    for _, row in mode_rows.iterrows():
                        mask |= (ls_filtered_df.index >= row['startDateTime']) & (ls_filtered_df.index < row['endDateTime'])
                    ls_filtered_df = ls_filtered_df[mask]
                    soo_mode_name = self.soo_dict[alarm_id]
                else:
                    print(f"Cannot check for {alarm_id} because there are no {alarm_id} periods in time frame.")
                    continue
            elif not ls_df.empty:
                # Filter out all date range rows from ls_filtered_df's indexes
                mask = pd.Series(True, index=ls_filtered_df.index)
                for _, row in ls_df.iterrows():
                    mask &= ~((ls_filtered_df.index >= row['startDateTime']) & (ls_filtered_df.index < row['endDateTime']))
                ls_filtered_df = ls_filtered_df[mask]

            for day in daily_df.index:
                next_day = day + pd.Timedelta(days=1)
                filtered_df = ls_filtered_df.loc[(ls_filtered_df.index >= day) & (ls_filtered_df.index < next_day)]
                id_group = non_pow_bounds_df[non_pow_bounds_df['alarm_code_id'] == alarm_id]
                on_t_codes = id_group[id_group['alarm_code_type'] == 'ON']
                off_t_codes = id_group[id_group['alarm_code_type'] == 'OFF']
                if len(on_t_codes) != 1 or len(off_t_codes) != 1:
                    raise Exception(f"Improper alarm codes for SOO changes with id {alarm_id}. Must have 1 ON and 1 OFF variable")
                on_t_var_name = on_t_codes.iloc[0]['variable_name']
                on_t_pretty_name = on_t_codes.iloc[0]['pretty_name']
                on_t_thresh = on_t_codes.iloc[0]['bound']
                off_t_var_name = off_t_codes.iloc[0]['variable_name']
                off_t_pretty_name = off_t_codes.iloc[0]['pretty_name']
                off_t_thresh = off_t_codes.iloc[0]['bound']
                if pow_var_name in filtered_df.columns:
                    power_below = filtered_df[pow_var_name] <= pow_thresh
                    power_above = filtered_df[pow_var_name] > pow_thresh

                    # Check all turn-on events
                    if on_t_var_name in filtered_df.columns:
                        power_turn_on = power_below.shift(1) & power_above
                        power_on_times = filtered_df.index[power_turn_on.fillna(False)]
                        # Check if temperature is within 5.0 of on_t_thresh at each turn-on moment
                        for power_time in power_on_times:
                            temp_at_turn_on = filtered_df.loc[power_time, on_t_var_name]
                            if abs(temp_at_turn_on - on_t_thresh) > 5.0:
                                self._add_an_alarm(power_time, power_time, on_t_var_name,
                                    f"Unexpected SOO change: during {soo_mode_name}, HP turned on at {power_time} but {on_t_pretty_name} was {temp_at_turn_on:.1f} F (setpoint at {on_t_thresh} F).",
                                    certainty="med")

                    # Check all turn-off events
                    if off_t_var_name in filtered_df.columns:
                        power_turn_off = power_above.shift(1) & power_below
                        power_off_times = filtered_df.index[power_turn_off.fillna(False)]
                        # Check if temperature is within 5.0 of off_t_thresh at each turn-off moment
                        for power_time in power_off_times:
                            temp_at_turn_off = filtered_df.loc[power_time, off_t_var_name]
                            if abs(temp_at_turn_off - off_t_thresh) > 5.0:
                                self._add_an_alarm(power_time, power_time, off_t_var_name,
                                    f"Unexpected SOO change: during {soo_mode_name}, HP turned off at {power_time} but {off_t_pretty_name} was {temp_at_turn_off:.1f} F (setpoint at {off_t_thresh} F).",
                                    certainty="med")
                                
    def _organize_alarm_codes(self, bounds_df : pd.DataFrame) -> list:
        alarm_code_parts = []
        for idx, row in bounds_df.iterrows():
            var_name_parts = row['variable_name'].split('_')
            if len(var_name_parts) <= 1:
                raise Exception(f"Improper variable name for '{row['variable_name']}', mustt be in form '[Unit Type]_[Element Identifier]' (e.g. 'Temp_HPWH' or 'PowerIn_SwingTank1').")
            if var_name_parts[0] == "PowerIn":
                alarm_code_parts.append(["PowerIn", "No ID"])
            elif var_name_parts[0] == "Temp":
                parts = row['alarm_codes'].split('_')
                if len(parts) == 2:
                    alarm_code_parts.append([parts[1], "No ID"])
                elif len(parts) == 3:
                    alarm_code_parts.append([parts[1], parts[2]])
                else:
                    raise Exception(f"improper {self.alarm_tag} alarm code format for {row['variable_name']}")
            else:
                raise Exception(f"{var_name_parts[0]} is not a proper unit type for SOOCHNG alarms.")

        return alarm_code_parts


import pandas as pd
import numpy as np
import datetime as datetime
from ecopipeline import ConfigManager
import re
import mysql.connector.errors as mysqlerrors
from datetime import timedelta
from ecopipeline.event_tracking.Alarm import Alarm

class TempRange(Alarm):
    """
    Detects when a temperature variable falls outside an acceptable range for a sustained period.
    An alarm triggers when the value stays above the high bound or below the low bound for
    fault_time consecutive minutes.

    Variable_Names.csv configuration:
      alarm_codes column: TMPRANG:low-high where low-high is the acceptable temperature range
        (e.g., TMPRANG:110-130 means temperature should stay between 110 and 130 degrees F).
      variable_name column: Must start with Temp_ (e.g., Temp_SwingTank1).
        Temp_[name] - Temperature variable to monitor. Bounds (low-high) come from alarm_codes.
            Alarm triggers if the temperature stays outside the [low, high] range for fault_time consecutive minutes.

    Parameters
    ----------
    default_high_temp : float
        Default high temperature bound when no range is specified in the alarm code (default 130).
        Temperature above this triggers an alarm.
    default_low_temp : float
        Default low temperature bound when no range is specified in the alarm code (default 115).
        Temperature below this triggers an alarm.
    fault_time : int
        Number of consecutive minutes that temperature must be outside the range before triggering an alarm (default 10).
    """
    def __init__(self, bounds_df : pd.DataFrame, default_high_temp : float = 130, default_low_temp : float = 115, fault_time : int = 10):
        alarm_tag = 'TMPRANG'
        type_default_dict = {'Temp': [default_low_temp, default_high_temp]}
        self.fault_time = fault_time
        super().__init__(bounds_df, alarm_tag,type_default_dict, two_part_tag = False, range_bounds=True)

    def specific_alarm_function(self, df: pd.DataFrame, daily_df : pd.DataFrame, config : ConfigManager):
        # Process each unique alarm_code_id
        for dhw_var in self.bounds_df['variable_name'].unique():
            for day in daily_df.index:
                next_day = day + pd.Timedelta(days=1)
                filtered_df = df.loc[(df.index >= day) & (df.index < next_day)]
                rows = self.bounds_df[self.bounds_df['variable_name'] == dhw_var]
                low_bound = rows.iloc[0]['bound']
                high_bound = rows.iloc[0]['bound2']
                pretty_name = rows.iloc[0]['pretty_name']

                if dhw_var in filtered_df.columns:
                    # Check if temp is above high bound or below low bound
                    out_of_range_mask = (filtered_df[dhw_var] > high_bound) | (filtered_df[dhw_var] < low_bound)

                    # Check for fault_time consecutive minutes
                    consecutive_condition = out_of_range_mask.rolling(window=self.fault_time).min() == 1
                    if consecutive_condition.any():
                        # Find all streaks of consecutive True values
                        group = (consecutive_condition != consecutive_condition.shift()).cumsum()

                        # Iterate through each streak and add an alarm for each
                        for group_id in consecutive_condition.groupby(group).first()[lambda x: x].index:
                            streak_indices = consecutive_condition[group == group_id].index
                            streak_length = len(streak_indices)

                            # Adjust start time because first (fault_time-1) minutes don't count in window
                            start_time = streak_indices[0] - pd.Timedelta(minutes=self.fault_time-1)
                            end_time = streak_indices[-1]
                            adjusted_streak_length = streak_length + self.fault_time - 1

                            self._add_an_alarm(start_time, end_time, dhw_var,
                                f"Temperature out of range: {pretty_name} was outside {low_bound}-{high_bound} F for {adjusted_streak_length} consecutive minutes starting at {start_time}.")


import pandas as pd
import numpy as np
import datetime as datetime
from ecopipeline import ConfigManager
import re
import mysql.connector.errors as mysqlerrors
from datetime import timedelta
from ecopipeline.event_tracking.Alarm import Alarm

class AbnormalCOP(Alarm):
    """
    Detects abnormal COP (Coefficient of Performance) values by checking if any COP variable
    falls outside its expected high/low bounds on a given day.

    Note: This alarm does not use the alarm_codes column. Variables are matched by their column name
    pattern in daily_df (must start with 'COP' or 'SystemCOP'), and bounds are read directly from the
    high_alarm and low_alarm columns in Variable_Names.csv.

    Variable_Names.csv columns:
    variable_name - Name of the COP variable to monitor (must match a column starting with 'COP' or 'SystemCOP' in daily_df).
    high_alarm - Upper bound for acceptable COP. Alarm triggers if daily COP exceeds this value. Default: 4.5.
    low_alarm - Lower bound for acceptable COP. Alarm triggers if daily COP falls below this value. Default: 0.
    pretty_name - (Optional) Display name for the variable in alarm messages. Defaults to variable_name.

    Parameters
    ----------
    default_high_bound : float
        Default upper COP bound when no high_alarm value is specified (default 4.5).
    default_low_bound : float
        Default lower COP bound when no low_alarm value is specified (default 0).
    """
    def __init__(self, bounds_df : pd.DataFrame, default_high_bound : float = 4.5, default_low_bound : float = 0):
        self.default_high_bound = default_high_bound
        self.default_low_bound = default_low_bound
        
        super().__init__(bounds_df, "ABNRMCP", {}, daily_only=True)

    def _process_bounds_df_alarm_codes(self, og_bounds_df : pd.DataFrame) -> pd.DataFrame:
        bounds_df = og_bounds_df.copy()
        if not "variable_name" in bounds_df.columns:
            raise Exception(f"variable_name is not present in Variable_Names.csv")
        if not 'pretty_name' in bounds_df.columns:
            bounds_df['pretty_name'] = bounds_df['variable_name']
        else:
            bounds_df['pretty_name'] = bounds_df['pretty_name'].fillna(bounds_df['variable_name'])
        if not 'high_alarm' in bounds_df.columns:
            bounds_df['high_alarm'] = self.default_high_bound
        else:
            bounds_df['high_alarm'] = bounds_df['high_alarm'].fillna(self.default_high_bound)
        if not 'low_alarm' in bounds_df.columns:
            bounds_df['low_alarm'] = self.default_low_bound
        else:
            bounds_df['low_alarm'] = bounds_df['low_alarm'].fillna(self.default_low_bound)

        bounds_df = bounds_df.loc[:, ["variable_name", "high_alarm", "low_alarm", "pretty_name"]]
        bounds_df.dropna(axis=0, thresh=2, inplace=True)
        bounds_df.set_index(['variable_name'], inplace=True)

        return bounds_df

    def specific_alarm_function(self, df: pd.DataFrame, daily_df : pd.DataFrame, config : ConfigManager):
        cop_pattern = re.compile(r'^(COP\w*|SystemCOP\w*)$')
        cop_columns = [col for col in daily_df.columns if re.match(cop_pattern, col)]

        if not daily_df.empty and len(cop_columns) > 0:
            for bound_var, bounds in self.bounds_df.iterrows():
                if bound_var in cop_columns:
                    for day, day_values in daily_df.iterrows():
                        if not day_values[bound_var] is None and (day_values[bound_var] > bounds['high_alarm'] or day_values[bound_var] < bounds['low_alarm']):
                            alarm_str = f"Unexpected COP Value detected: {bounds['pretty_name']} = {round(day_values[bound_var],2)}"
                            self._add_an_alarm(day, day + timedelta(1), bound_var, alarm_str, add_one_minute_to_end=False)


import pandas as pd
import numpy as np
import datetime as datetime
from ecopipeline import ConfigManager
import re
import mysql.connector.errors as mysqlerrors
from datetime import timedelta
from ecopipeline.event_tracking.Alarm import Alarm

class PowerRatio(Alarm):
    """
    Detects when power variables fall outside their expected ratio of a group total over a rolling period.
    Variables are grouped by element ID derived from their variable_name, and each variable's contribution
    is checked against its expected low-high percentage range.

    Variable_Names.csv configuration:
      alarm_codes column: POWRRAT:low-high where low-high is the acceptable percentage range
        (e.g., POWRRAT:60-80 means the variable should account for 60-80% of its group total).
        Bounds (low-high) come from alarm_codes and are not part of the variable_name.
      variable_name column: Must start with PowerIn_. Determines the grouping of variables:
        PowerIn_Total - Total system power. Used as the denominator for the 'Total' group.
        PowerIn_[name containing '_HPWH'] - Grouped under the 'HPWH' element ID. The sum of all HPWH
            variables is the group denominator; each variable's ratio is checked against its range.
        PowerIn_[other] - All other PowerIn variables are grouped under the 'Total' element ID and their
            ratios are calculated against the PowerIn_Total variable.

    Parameters
    ----------
    day_table_name : str
        Name of the site's daily aggregate value table in the database for fetching historical data.
    ratio_period_days : int
        Number of days to use for the rolling power ratio calculation (default 7). Each block sums
        values over this many days before calculating ratios.
    """
    def __init__(self, bounds_df : pd.DataFrame, day_table_name : str, ratio_period_days : int = 7):
        alarm_tag = 'POWRRAT'
        type_default_dict = {}
        self.ratio_period_days = ratio_period_days
        self.day_table_name = day_table_name # TODO this could be a security issue. Swap it for config manager
        super().__init__(bounds_df, alarm_tag,type_default_dict, two_part_tag = False, range_bounds=True, daily_only=True)

    def specific_alarm_function(self, df: pd.DataFrame, daily_df : pd.DataFrame, config : ConfigManager):
        daily_df_copy = daily_df.copy()
        if self.ratio_period_days > 1:
            daily_df_copy = self._append_previous_days_to_df(daily_df_copy, config, self.ratio_period_days, self.day_table_name)
        elif self.ratio_period_days < 1:
            print("power ratio alarm period, ratio_period_days, must be more than 1")
            return

        # Create blocks of ratio_period_days
        blocks_df = self._create_period_blocks(daily_df_copy)

        if blocks_df.empty:
            print("No complete blocks available for analysis")
            return

        for alarm_id in self.bounds_df['alarm_code_id'].unique():
            # Calculate total for each block
            var_list = self.bounds_df[self.bounds_df['alarm_code_id'] == alarm_id]['variable_name'].unique()
            for var in var_list:
                if var not in blocks_df.columns:
                    blocks_df[var] = 0.0
            if alarm_id == 'Total':
                tp_codes = self.bounds_df[self.bounds_df['alarm_code_type'] == 'PowerIn_Total']
                if len(tp_codes) != 1:
                    raise Exception(f"POWRRAT Error: There must be exactly one Total Power variable. Found {len(tp_codes)}")
                if tp_codes.iloc[0]['variable_name'] in blocks_df.columns:
                    blocks_df[alarm_id] = blocks_df[tp_codes.iloc[0]['variable_name']]
                    var_list = var_list[var_list != tp_codes.iloc[0]['variable_name']]
                else:
                   raise Exception(f"POWRRAT Error: PowerIn_Total variable missing from total power ratio") 
            else:
                blocks_df[alarm_id] = blocks_df[var_list].sum(axis=1)

            for variable in var_list:
                # Calculate ratio for each block
                blocks_df[f"{variable}_{alarm_id}"] = (blocks_df[variable]/blocks_df[alarm_id]) * 100

                # Get bounds from bounds_df for this variable and alarm_id
                var_row = self.bounds_df[(self.bounds_df['variable_name'] == variable) & (self.bounds_df['alarm_code_id'] == alarm_id)]
                if var_row.empty:
                    continue
                low_bound = var_row.iloc[0]['bound']
                high_bound = var_row.iloc[0]['bound2']
                pretty_name = var_row.iloc[0]['pretty_name']

                alarm_blocks_df = blocks_df.loc[(blocks_df[f"{variable}_{alarm_id}"] < low_bound) | (blocks_df[f"{variable}_{alarm_id}"] > high_bound)]
                if not alarm_blocks_df.empty:
                    for block_end_date, values in alarm_blocks_df.iterrows():
                        block_start_date = block_end_date - timedelta(days=self.ratio_period_days - 1)
                        actual_ratio = values[f'{variable}_{alarm_id}']
                        self._add_an_alarm(block_start_date, block_end_date + timedelta(1), variable,
                            f"Power ratio alarm ({self.ratio_period_days}-day block ending {block_end_date.strftime('%Y-%m-%d')}): {pretty_name} accounted for {actual_ratio:.1f}% of {alarm_id} energy use. {low_bound:.1f}-{high_bound:.1f}% expected.", add_one_minute_to_end=False) 
    
    def _create_period_blocks(self, daily_df: pd.DataFrame, verbose: bool = False) -> pd.DataFrame:
        """
        Create blocks of ratio_period_days by summing values within each block.
        Each block will be represented by its end date.
        """
        if len(daily_df) < self.ratio_period_days:
            if verbose:
                print(f"Not enough data for {self.ratio_period_days}-day blocks. Need at least {self.ratio_period_days} days but only have {len(daily_df)} days in data")
            return pd.DataFrame()
        
        blocks = []
        block_dates = []
        
        # Create blocks by summing consecutive groups of ratio_period_days
        for i in range(self.ratio_period_days - 1, len(daily_df)):
            start_idx = i - self.ratio_period_days + 1
            end_idx = i + 1
            
            block_data = daily_df.iloc[start_idx:end_idx].sum()
            blocks.append(block_data)
            # Use the end date of the block as the identifier
            block_dates.append(daily_df.index[i])
        
        if not blocks:
            return pd.DataFrame()
        
        blocks_df = pd.DataFrame(blocks, index=block_dates)
        
        if verbose:
            print(f"Created {len(blocks_df)} blocks of {self.ratio_period_days} days each")
            print(f"Block date range: {blocks_df.index.min()} to {blocks_df.index.max()}")
        
        return blocks_df
    
    def _organize_alarm_codes(self, bounds_df : pd.DataFrame) -> list:
        alarm_code_parts = []
        seen_total_power = False
        for idx, row in bounds_df.iterrows():
            element_id = 'Total'
            parts = row['variable_name'].split('_')
            if len(parts) <= 1:
                raise Exception(f"Improper variable name for '{row['variable_name']}', must be in form '[Unit Type]_[Element Identifier]' (e.g. 'Temp_HPWH' or 'PowerIn_SwingTank1').")
            elif parts[0] != "PowerIn":
                raise Exception(f"Error in POWRRAT: {row['variable_name']} is not designated as a power variable (must begin with 'PowerIn_').")
            if parts[0] == "PowerIn" and parts[1] == "Total":
                # total power is own catagory
                if seen_total_power:
                    raise Exception(f"Multiple instances of PowerIn_Total seen for alarm code {self.alarm_tag}. There may only be one variable that starts with 'PowerIn_Total'. This should be total system power.")
                alarm_code_parts.append(["PowerIn_Total", element_id])
                seen_total_power = True
            else:
                if "_HPWH" in row['variable_name']:
                    element_id = "HPWH"
                    
                alarm_code_parts.append([parts[0], element_id])

        return alarm_code_parts